from typing import Dict, List, Union

JSONPrimitive = Union[str, int, None, float]
JSONValue = Union[JSONPrimitive, "JSONObject", List["JSONValue"]]
JSONObject = Dict[str, JSONValue]

ProcessVariable = Dict[str, JSONPrimitive]  # {"type": "Integer", "value": 42}
ProcessVariables = Dict[str, ProcessVariable]



import json
from datetime import date, datetime
from typing import Any, Dict, List, Union

import inflection
from dateutil import parser

from .types import ProcessVariable


def noop(val):
    return val


def underscoreize(data: Union[List, Dict, str, None]) -> Union[List, Dict, str, None]:
    if isinstance(data, list):
        return [underscoreize(item) for item in data]

    if isinstance(data, dict):
        new_data = {}
        for key, value in data.items():
            new_key = inflection.underscore(key)
            # variables are dynamic names, can't make assumptions!
            if key == "variables":
                new_data[new_key] = value
            else:
                new_data[new_key] = underscoreize(value)
        return new_data

    return data


TYPE_MAP = {
    bool: ("Boolean", noop),
    date: (
        "String",
        lambda d: d.isoformat(),
    ),  # Date object requires time information, which we don't have
    datetime: ("Date", lambda d: d.isoformat()),
    int: ("Integer", noop),
    float: ("Double", noop),
    str: ("String", noop),
    type(None): ("Null", noop),
    dict: ("Json", json.dumps),
    list: ("Json", json.dumps),
}


REVERSE_TYPE_MAP = {
    "date": parser.parse,
    "json": json.loads,
    "integer": int,
    "short": int,
    "long": int,
}


def serialize_variable(value: Any) -> ProcessVariable:
    val_type = type(value)
    if val_type not in TYPE_MAP:
        raise NotImplementedError(f"Type {val_type} is not implemented yet")

    type_name, converter = TYPE_MAP[val_type]
    return {"type": type_name, "value": converter(value)}


def deserialize_variable(variable: ProcessVariable) -> Any:
    var_type = variable.get("type", "String")
    converter = REVERSE_TYPE_MAP.get(var_type.lower())
    if converter:
        value = converter(variable["value"])
    else:
        value = variable["value"]  # a JSON primitive that maps to proper python objects

    return value



"""
Public Python API to interact with Activiti.
"""
import uuid
from typing import Dict, List, Optional, Union

from .camunda_models import ProcessDefinition, factory
from .client import get_client
from .types import JSONObject, ProcessVariable, ProcessVariables
from .utils import deserialize_variable


def get_process_definitions() -> List[ProcessDefinition]:
    client = get_client()
    response = client.get("process-definition", {"sortBy": "name", "sortOrder": "asc"})
    return factory(ProcessDefinition, response)


def get_start_form_variables(
    process_key: Optional[str] = None, process_id: Optional[str] = None
) -> Dict[str, JSONObject]:
    """
    Get the start form variables from the Camunda process.

    If defaults are configured in the Camunda process, these will be returned as value.

    A process ID is more specific than a process key, so if both are provided, the
    process ID will be used.
    """
    if not (process_key or process_id):
        raise ValueError("Provide a process key or process ID")

    if process_id:
        endpoint = f"process-definition/{process_id}/form-variables"
    else:
        endpoint = f"process-definition/key/{process_key}/form-variables"

    # TODO: do any necessary type casting
    client = get_client()
    variables = client.get(endpoint, underscoreize=False)

    return variables


def get_process_instance_variable(
    instance_id: Union[uuid.UUID, str], name: str
) -> ProcessVariable:
    client = get_client()

    response_data = client.get(
        f"process-instance/{instance_id}/variables/{name}",
        params={"deserializeValues": "false"},
        underscoreize=False,
    )
    return deserialize_variable(response_data)


def get_all_process_instance_variables(
    instance_id: Union[uuid.UUID, str]
) -> ProcessVariables:
    client = get_client()

    response_data = client.get(
        f"process-instance/{instance_id}/variables",
        params={"deserializeValues": "false"},
        underscoreize=False,
    )
    variables = {
        name: deserialize_variable(variable) for name, variable in response_data.items()
    }
    return variables



# Generated by Django 2.2.9 on 2020-01-13 16:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_camunda", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="camundaconfig",
            name="auth_header",
            field=models.TextField(
                blank=True,
                help_text="HTTP Authorization header value, required if the API is not open.",
                verbose_name="authorization header",
            ),
        )
    ]



"""
Implements a camunda client.
"""
import logging
from typing import Any, Optional, Union
from urllib.parse import urljoin

from django.conf import settings
from django.utils.module_loading import import_string

import requests

from .models import CamundaConfig
from .utils import underscoreize

logger = logging.getLogger(__name__)


def get_client_class() -> type:
    client_class = getattr(
        settings, "CAMUNDA_CLIENT_CLASS", "django_camunda.client.Camunda"
    )
    return import_string(client_class)


def get_client(*args, **kwargs):
    client_class = get_client_class()
    return client_class(*args, **kwargs)


class Camunda:
    def __init__(self, config: Optional[CamundaConfig] = None):
        self.config = config or CamundaConfig.get_solo()
        self.root_url = self.config.api_root

    @property
    def auth(self) -> dict:
        if not self.config.auth_header:
            return {}
        return {"Authorization": self.config.auth_header}

    def request(self, path: str, method="GET", *args, **kwargs):
        assert not path.startswith("/"), "Provide relative API paths"
        url = urljoin(self.root_url, path)

        do_underscoreize = kwargs.pop("underscoreize", True)

        # add the API headers, so that Camunda can use the tokens. Essentially
        # we're forwarding Auth
        headers = kwargs.pop("headers", {})
        headers.update(self.auth)
        headers.update(self.get_extra_headers(headers))
        kwargs["headers"] = headers

        json = kwargs.get("json")
        if json:
            self.preprocess_json(json)

        _ref = self.before_request(method, url, *args, **kwargs)

        response = requests.request(method, url, *args, **kwargs)
        response_data = None

        try:
            response.raise_for_status()
            if response.content:
                # json is the default Content-Type
                content_type = response.headers.get("Content-Type", "application/json")
                if content_type.startswith("application/json"):
                    response_data = response.json()

                    if isinstance(response_data, (dict, list)):
                        self.postprocess_response_data(response_data)

                    if do_underscoreize:
                        response_data = underscoreize(response_data)
                else:
                    # binary content
                    response_data = response.content

            return response_data
        except Exception:
            try:
                # see if we can grab any extra output
                response_data = response.json()
            except Exception:
                pass
            logger.exception("Error: %r", response_data)
            raise
        finally:
            self.after_request(_ref, response, response_data)

    def get(self, path: str, params=None, *args, **kwargs):
        return self.request(path, method="GET", params=params, *args, **kwargs)

    def post(self, path: str, data=None, json=None, *args, **kwargs):
        return self.request(path, method="POST", data=data, json=json, *args, **kwargs)

    def put(self, path: str, data=None, *args, **kwargs):
        return self.request(path, method="PUT", data=data, *args, **kwargs)

    def patch(self, path: str, data=None, *args, **kwargs):
        return self.request(path, method="PATCH", data=data, *args, **kwargs)

    def delete(self, path: str, *args, **kwargs):
        return self.request(path, method="DELETE", *args, **kwargs)

    # HOOKS for subclasses

    def before_request(self, method: str, url: str, *args, **kwargs) -> Any:
        pass

    def after_request(self, ref: Any, response, response_data) -> None:
        pass

    def get_extra_headers(self, headers: dict) -> dict:
        return {}

    def preprocess_json(self, json: dict) -> None:
        pass

    def postprocess_response_data(self, data: Union[list, dict]) -> None:
        pass



from django.contrib import admin
from django.contrib.admin import widgets
from django.utils.translation import gettext_lazy as _

from solo.admin import SingletonModelAdmin

from .fields import ProcessDefinitionField
from .forms import CamundaConfigForm, get_process_definition_choices
from .models import CamundaConfig


@admin.register(CamundaConfig)
class CamundaConfigAdmin(SingletonModelAdmin):
    form = CamundaConfigForm
    fieldsets = (
        (None, {"fields": ("root_url", "rest_api_path")}),
        (
            _("Auth"),
            {
                "fields": (
                    ("basic_auth_username", "basic_auth_password"),
                    ("auth_header",),
                )
            },
        ),
    )
    readonly_fields = ("auth_header",)


class CamundaFieldsMixin:
    def formfield_for_dbfield(self, db_field, request, **kwargs):
        if isinstance(db_field, ProcessDefinitionField):
            kwargs.update(
                {
                    "widget": widgets.AdminRadioSelect(),
                    "choices": get_process_definition_choices(),
                }
            )

        return super().formfield_for_dbfield(db_field, request, **kwargs)



import uuid
from dataclasses import dataclass
from datetime import date, datetime
from typing import Any, List, Optional, Union

from dateutil.parser import parse

from .types import JSONObject


def noop(value: Any) -> Any:
    return value


CONVERTERS = {
    type(None): lambda x: None,
    str: noop,
    int: noop,
    float: noop,
    dict: noop,  # TODO: recurse?
    uuid.UUID: lambda value: uuid.UUID(value),
    datetime: parse,
    date: date.fromisoformat,
}


class Model:
    def __post_init__(self):
        self._type_cast()

    def _type_cast(self):
        for attr, typehint in self.__annotations__.items():
            value = getattr(self, attr)

            if typehint is None:
                typehint = type(None)

            # support for Optional
            if hasattr(typehint, "__origin__") and typehint.__origin__ is Union:
                typehint = typehint.__args__

                if value is None:
                    continue

                # Optional is ONE type combined with None
                typehint = next(t for t in typehint if t is not None)

            if isinstance(value, typehint):
                continue

            converter = CONVERTERS[typehint]
            setattr(self, attr, converter(value))


def factory(model: type, data: Union[JSONObject, List[JSONObject]]) -> type:
    _is_collection = isinstance(data, list)

    known_kwargs = list(model.__annotations__.keys())

    def _normalize(kwargs: dict):
        to_keep = {key: value for key, value in kwargs.items() if key in known_kwargs}
        return to_keep

    if not _is_collection:
        data = [data]

    instances = [model(**_normalize(_raw)) for _raw in data]

    if not _is_collection:
        instances = instances[0]
    return instances


@dataclass
class Task(Model):
    id: uuid.UUID
    name: str
    assignee: Optional[str]
    created: datetime
    due: Optional[datetime]
    follow_up: Optional[str]
    delegation_state: Optional[str]
    description: Optional[str]
    execution_id: str
    owner: Optional[str]
    parent_task_id: Optional[uuid.UUID]
    priority: int
    process_definition_id: str
    process_instance_id: uuid.UUID
    task_definition_key: str
    case_execution_id: Optional[str]
    case_instance_id: Optional[str]
    case_definition_id: Optional[str]
    suspended: bool
    form_key: Optional[str]
    tenant_id: Optional[str]


@dataclass
class ProcessDefinition(Model):
    id: str
    key: str
    name: str
    category: str
    version: int
    deployment_id: uuid.UUID
    resource: str  # filename
    startable_in_tasklist: bool
    suspended: bool
    description: Optional[str] = None
    tenant_id: Optional[str] = None
    version_tag: Optional[str] = None  # unsure
    diagram: Optional[str] = None  # unsure
    history_time_to_live: Optional[str] = None  # unsure



from django.core.management import BaseCommand

from ...client import get_client


class Command(BaseCommand):
    help = "Delete process instances by ID"

    def add_arguments(self, parser):
        parser.add_argument("id", nargs="+", help="Process instance IDs to delete")

    def handle(self, **options):
        client = get_client()

        for _id in options["id"]:
            client.delete(f"process-instance/{_id}")
            self.stdout.write(f"Deleted {_id}")
